/*
 * SPDX-FileCopyrightText: 2018-2018 CSSlayer <wengxt@gmail.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 *
 */
#include "pipelinejob.h"

namespace fcitx {

PipelineJob::PipelineJob(QObject *parent) : QObject(parent) {}

} // namespace fcitx
